from __future__ import division
from scipy import sparse
from scipy.signal import butter, lfilter
from scipy.stats import norm
from sklearn import base
from statsmodels.distributions.empirical_distribution import ECDF
import logging
import numpy as np
import pandas as pd


NAN_INT = 7535805


class Normalizer(base.BaseEstimator):
    """Normalizer that transforms numerical columns into normal distribution.

    Attributes:
        ecdfs (list of empirical CDF): empirical CDFs for columns
    """

    def fit(self, X, y=None):
        self.ecdfs = [None] * X.shape[1]

        for col in range(X.shape[1]):
            self.ecdfs[col] = ECDF(X[:, col])

        return self

    def transform(self, X):
        """Normalize numerical columns.
        
        Args:
            X (numpy.array) : numerical columns to normalize

        Returns:
            X (numpy.array): normalized numerical columns
        """

        for col in range(X.shape[1]):
            X[:, col] = self._transform_col(X[:, col], col)
            
        return X

    def fit_transform(self, X, y=None):
        """Normalize numerical columns.
        
        Args:
            X (numpy.array) : numerical columns to normalize

        Returns:
            X (numpy.array): normalized numerical columns
        """

        self.ecdfs = [None] * X.shape[1]

        for col in range(X.shape[1]):
            self.ecdfs[col] = ECDF(X[:, col])
            X[:, col] = self._transform_col(X[:, col], col)

        return X

    def _transform_col(self, x, col):
        """Normalize one numerical column.
        
        Args:
            x (numpy.array): a numerical column to normalize
            col (int): column index

        Returns:
            A normalized feature vector.
        """

        return norm.ppf(self.ecdfs[col](x) * .998 + .001)


class LabelEncoder(base.BaseEstimator):
    """Label Encoder that groups infrequent values into one label.

    Attributes:
        min_obs (int): minimum number of observation to assign a label.
        label_encoders (list of dict): label encoders for columns
        label_maxes (list of int): maximum of labels for columns
    """

    def __init__(self, min_obs=10):
        """Initialize the OneHotEncoder class object.

        Args:
            min_obs (int): minimum number of observation to assign a label.
        """

        self.min_obs = min_obs

    def __repr__(self):
        return ('LabelEncoder(min_obs={})').format(self.min_obs)

    def _get_label_encoder_and_max(self, x):
        """Return a mapping from values and its maximum of a column to integer labels.

        Args:
            x (numpy.array): a categorical column to encode.

        Returns:
            label_encoder (dict): mapping from values of features to integers
            max_label (int): maximum label
        """

        # NaN cannot be used as a key for dict. So replace it with a random integer.
        x[pd.isnull(x)] = NAN_INT

        # count each unique value
        label_count = {}
        for label in x:
            try:
                label_count[label] += 1
            except KeyError:
                label_count[label] = 1

        # add unique values appearing more than min_obs to the encoder.
        label_encoder = {}
        label_index = 1
        labels_not_encoded = 0
        for label in label_count.keys():
            if label_count[label] >= self.min_obs:
                label_encoder[label] = label_index
                label_index += 1
            else:
                labels_not_encoded += 1

        max_label = label_index - 1

        # if every label is encoded, then replace the maximum label with 0 so
        # that total number of labels encoded is (# of total labels - 1).
        if labels_not_encoded == 0:
            for label in label_encoder:
                # find the label with the maximum encoded value
                if label_encoder[label] == max_label:
                    # set the value of the label to 0 and decrease the maximum
                    # by 1.
                    label_encoder[label] = 0
                    max_label -= 1
                    break

        return label_encoder, max_label

    def _transform_col(self, x, col):
        """Encode one categorical column into labels.

        Args:
            x (numpy.array): a categorical column to encode
            col (int): column index

        Returns:
            x (numpy.array): a column with labels.
        """

        label_encoder = self.label_encoders[col]

        # replace NaNs with the pre-defined random integer
        x[pd.isnull(x)] = NAN_INT

        labels = np.zeros((x.shape[0], ), dtype=np.int64)
        for label in label_encoder:
            labels[x == label] = label_encoder[label]

        return labels

    def fit(self, X, y=None):
        self.label_encoders = [None] * X.shape[1]
        self.label_maxes = [None] * X.shape[1]

        for col in range(X.shape[1]):
            self.label_encoders[col], self.label_maxes[col] = \
                self._get_label_encoder_and_max(X[:, col])

        return self

    def transform(self, X):
        """Encode categorical columns into sparse matrix with one-hot-encoding.

        Args:
            X (numpy.array): categorical columns to encode

        Returns:
            X (numpy.array): label encoded columns
        """

        for col in range(X.shape[1]):
            X[:, col] = self._transform_col(X[:, col], col)

        return X

    def fit_transform(self, X, y=None):
        """Encode categorical columns into label encoded columns

        Args:
            X (numpy.array): categorical columns to encode

        Returns:
            X (numpy.array): label encoded columns
        """

        self.label_encoders = [None] * X.shape[1]
        self.label_maxes = [None] * X.shape[1]

        for col in range(X.shape[1]):
            self.label_encoders[col], self.label_maxes[col] = \
                self._get_label_encoder_and_max(X[:, col])

            X[:, col] = self._transform_col(X[:, col], col)

        return X


class OneHotEncoder(base.BaseEstimator):
    """One-Hot-Encoder that groups infrequent values into one dummy variable.

    Attributes:
        min_obs (int): minimum number of observation to create a dummy variable
        label_encoders (list of (dict, int)): label encoders and their maximums
                                              for columns
    """

    def __init__(self, min_obs=10):
        """Initialize the OneHotEncoder class object.

        Args:
            min_obs (int): minimum number of observation to create a dummy variable
            label_encoder (LabelEncoder): LabelEncoder that transofrm
        """

        self.min_obs = min_obs
        self.label_encoder = LabelEncoder(min_obs)

    def __repr__(self):
        return ('OneHotEncoder(min_obs={})').format(self.min_obs)

    def _transform_col(self, x, col):
        """Encode one categorical column into sparse matrix with one-hot-encoding.

        Args:
            x (numpy.array): a categorical column to encode
            col (int): column index

        Returns:
            X (scipy.sparse.coo_matrix): sparse matrix encoding a categorical
                                         variable into dummy variables
        """

        labels = self.label_encoder._transform_col(x, col)
        label_max = self.label_encoder.label_maxes[col]

        # build row and column index for non-zero values of a sparse matrix
        index = np.array(range(len(labels)))
        i = index[labels > 0]
        j = labels[labels > 0] - 1  # column index starts from 0

        if len(i) > 0:
            return sparse.coo_matrix((np.ones_like(i), (i, j)),
                                     shape=(x.shape[0], label_max))
        else:
            # if there is no non-zero value, return no matrix
            return None

    def fit(self, X, y=None):
        self.label_encoder.fit(X)

        return self

    def transform(self, X):
        """Encode categorical columns into sparse matrix with one-hot-encoding.

        Args:
            X (numpy.array): categorical columns to encode

        Returns:
            X_new (scipy.sparse.coo_matrix): sparse matrix encoding categorical
                                             variables into dummy variables
        """

        for col in range(X.shape[1]):
            X_col = self._transform_col(X[:, col], col)
            if X_col is not None:
                if col == 0:
                    X_new = X_col
                else:
                    X_new = sparse.hstack((X_new, X_col))

            logging.debug('{} --> {} features'.format(
                col, self.label_encoder.label_maxes[col])
            )

        return X_new

    def fit_transform(self, X, y=None):
        """Encode categorical columns into sparse matrix with one-hot-encoding.

        Args:
            X (numpy.array): categorical columns to encode

        Returns:
            sparse matrix encoding categorical variables into dummy variables
        """

        self.label_encoder.fit(X)

        return self.transform(X)


class BandpassFilter(base.BaseEstimator):

    def __init__(self, fs=10., lowcut=.5, highcut=3., order=3):
        self.fs = 10.
        self.lowcut = .5
        self.highcut = 3.
        self.order = 3
        b, a = _butter_bandpass()
        self.a = a
        self.b = b

    def _butter_bandpass(self):
        nyq = .5 * self.fs
        low = lowcut / nyq
        high = highcut / nyq
        b, a = butter(self.order, [low, high], btype='band')

        return b, a

    def _butter_bandpass_filter(self, x):
        return lfilter(self.b, self.a, x)

    def fit(self, X):
        return self

    def transform(self, X, y=None):
        for col in range(X.shape[1]):
            X[:, col] = self._butter_bandpass_filter(X[:, col])

        return X



__version__ = '0.5.0'
__all__ = ['const',
           'data_io',
           'feature_selection',
           'metrics',
           'model',
           'online_model',
           'preprocessing',
           'util']



from __future__ import division
from sklearn.metrics import mean_squared_error

import numpy as np


def rmse(y, p):
    """Root Mean Squared Error (RMSE).
    
    Args:
        y (numpy.array): target
        p (numpy.array): prediction

    Returns:
        e (numpy.float64): RMSE
    """

    # check and get number of samples
    assert y.shape == p.shape

    return np.sqrt(mean_squared_error(y, p))


def gini(y, p):
    """Normalized Gini Coefficient.
    
    Args:
        y (numpy.array): target
        p (numpy.array): prediction

    Returns:
        e (numpy.float64): normalized Gini coefficient
    """

    # check and get number of samples
    assert y.shape == p.shape

    n_samples = y.shape[0]
    
    # sort rows on prediction column
    # (from largest to smallest)
    arr = np.array([y, p]).transpose()
    true_order = arr[arr[:,0].argsort()][::-1,0]
    pred_order = arr[arr[:,1].argsort()][::-1,0]
    
    # get Lorenz curves
    l_true = np.cumsum(true_order) / np.sum(true_order)
    l_pred = np.cumsum(pred_order) / np.sum(pred_order)
    l_ones = np.linspace(1/n_samples, 1, n_samples)
    
    # get Gini coefficients (area between curves)
    g_true = np.sum(l_ones - l_true)
    g_pred = np.sum(l_ones - l_pred)
    
    # normalize to true Gini coefficient
    return g_pred / g_true



# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True
from __future__ import division
import numpy as np

cimport cython
from libc.math cimport sqrt, abs
from ..util cimport sigm
cimport numpy as np


np.import_array()


cdef class SGD:
    """Simple online learner using a hasing trick.

    Attributes:
        epoch (int): number of epochs
        n (int): number of features after hashing trick
        a (double): initial learning rate
        l1 (double): L1 regularization parameter
        l2 (double): L2 regularization parameter
        w (array of double): feature weights
        c (array of double): counters for weights
        interaction (boolean): whether to use 2nd order interaction or not
    """
    cdef unsigned int epoch
    cdef unsigned int n
    cdef double a
    cdef double l1
    cdef double l2
    cdef double[:] w
    cdef double[:] c
    cdef bint interaction

    def __init__(self,
                 double a=0.01,
                 double l1=0.0,
                 double l2=0.0,
                 unsigned int n=2**20,
                 unsigned int epoch=10,
                 bint interaction=True):
        """Initialize the SGD class object.

        Args:
            epoch (int): number of epochs
            n (int): number of features after hashing trick
            a (double): initial learning rate
            l1 (double): L1 regularization parameter
            l2 (double): L2 regularization parameter
            w (array of double): feature weights
            c (array of double): counters for weights
            interaction (boolean): whether to use 2nd order interaction or not
        """

        self.epoch = epoch
        self.n = n      # # of features
        self.a = a      # learning rate
        self.l1 = l1
        self.l2 = l2

        # initialize weights and counts
        self.w = np.zeros((self.n + 1,), dtype=np.float64)
        self.c = np.zeros((self.n + 1,), dtype=np.float64)
        self.interaction = interaction

    def __repr__(self):
        return ('SGD(a={}, l1={}, l2={}, n={}, epoch={}, interaction={})').format(
            self.a, self.l1, self.l2, self.n, self.epoch, self.interaction
        )

    def _indices(self, list x):
        cdef unsigned int index
        cdef int l
        cdef int i
        cdef int j

        yield self.n

        for index in x:
            yield abs(hash(index)) % self.n

        if self.interaction:
            l = len(x)
            x = sorted(x)
            for i in xrange(l):
                for j in xrange(i + 1, l):
                    yield abs(hash('{}_{}'.format(x[i], x[j]))) % self.n

    def read_sparse(self, path):
        """Apply hashing trick to the libsvm format sparse file.

        Args:
            path (str): a file path to the libsvm format sparse file

        Yields:
            x (list of int): a list of index of non-zero features
            y (int): target value
        """
        for line in open(path):
            xs = line.rstrip().split(' ')

            y = int(xs[0])
            x = []
            for item in xs[1:]:
                index, _ = item.split(':')
                x.append(abs(hash(index)) % self.n)

            yield x, y

    def fit(self, X, y):
        """Update the model with a sparse input feature matrix and its targets.

        Args:
            X (scipy.sparse.csr_matrix): a list of (index, value) of non-zero features
            y (numpy.array): targets

        Returns:
            updated model weights and counts
        """
        for epoch in range(self.epoch):
            for row in range(X.shape[0]):
                x = list(X[row].indices)
                self.update_one(x, self.predict_one(x) - y[row])

    def predict(self, X):
        """Predict for a sparse matrix X.

        Args:
            X (scipy.sparse.csr_matrix): a sparse matrix for input features

        Returns:
            p (numpy.array): predictions for input features
        """
        p = np.zeros((X.shape[0], ), dtype=np.float64)
        for row in range(X.shape[0]):
            p[row] = self.predict_one(list(X[row].indices))

        return p

    def predict_one(self, list x):
        """Predict for features.

        Args:
            x (list of int): a list of index of non-zero features

        Returns:
            p (double): a prediction for input features
        """
        cdef int i
        cdef double wTx

        wTx = 0.
        for i in self._indices(x):
            wTx += self.w[i]

        return sigm(wTx)

    def update_one(self, list x, double e):
        """Update the model.

        Args:
            x (list of int): a list of index of non-zero features
            e (double): error between the prediction of the model and target

        Returns:
            updates model weights and counts
        """
        cdef int i
        cdef double g2

        g2 = e * e
        for i in self._indices(x):
            self.w[i] -= (e +
                          (self.l1 if self.w[i] >= 0. else -self.l1) +
                          self.l2 * self.w[i]) * self.a / (sqrt(self.c[i]) + 1)
            self.c[i] += g2



from __future__ import division
from sklearn.metrics import roc_auc_score as auc
from sklearn.metrics import log_loss


def logloss(y, p):
    """Bounded log loss error.
    
    Args:
        y (numpy.array): target
        p (numpy.array): prediction

    Returns:
        bounded log loss error
    """

    p[p < 1e-15] = 1e-15
    p[p > 1 - 1e-15] = 1 - 1e-15
    return log_loss(y, p)



from scipy import sparse
from sklearn import base
import numpy as np


"""
Code below was originally written by Baris Umog (https://www.kaggle.com/barisumog).
"""
class DropInactive(base.BaseEstimator, base.TransformerMixin):

    def __init__(self, lowest=25):
        self.lowest = lowest

    def fit(self, X, y=None):
        x = (X > 0.0).astype(bool)
        s = np.array(x.sum(axis=0)).flatten()
        self.mask = (s >= self.lowest)
        return self

    def transform(self, X):
        print(self.mask.sum())
        return X[:, self.mask]


class DropLowInfo(base.BaseEstimator, base.TransformerMixin):

    def __init__(self, margin=0.02, weighted=True):
        self.margin = margin
        self.weighted = weighted

    def fit(self, X, y=None):
        mean = y.mean()
        lower = mean - self.margin
        upper = mean + self.margin
        ys = sparse.csc_matrix(y[:, np.newaxis])
        if self.weighted:
            x = X.multiply(ys).sum(axis=0)
            x = x / X.sum(axis=0)
        else:
            x = (X > 0)
            s = x.sum(axis=0)
            x = x.multiply(ys).sum(axis=0) / s
        x = np.array(x).flatten().astype('f4')
        mask1 = (x < lower)
        mask2 = (x > upper)
        self.mask = (mask1 + mask2).astype(bool)
        return self

    def transform(self, X):
        print(self.mask.sum())
        return X[:, self.mask]



# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True
from __future__ import division
from scipy import sparse

import logging
import numpy as np

cimport cython
from libc.math cimport exp, log
cimport numpy as np


np.import_array()


cdef double sigm(double x):
    """Bounded sigmoid function."""
    return 1 / (1 + exp(-fmax(fmin(x, 20.0), -20.0)))


def get_downsampled_index(n, rate=0.):
    """Return the index that downsamples a vector x by the rate."""

    return np.random.choice(range(n), int(n * rate), replace=False)


def get_downsampled_index0(x, rate=0., threshold=0.):
    """Return the index that downsamples 0s of a vector x by the rate."""

    idx1 = np.where(x > threshold)[0]
    idx0 = np.where(x <= threshold)[0]
    idx0_down = np.random.choice(idx0, int(len(idx0) * rate), replace=False)

    idx = list(idx0_down) + list(idx1)
    np.random.shuffle(idx)

    return idx


def set_column_width(X, n_col):
    """Set the column width of a matrix X to n_col."""

    if X.shape[1] < n_col:
        if sparse.issparse(X):
            X = sparse.hstack((X, np.zeros((X.shape[0], n_col - X.shape[1]))))
            X = X.tocsr()
        else:
            X = np.hstack((X, np.zeros((X.shape[0], n_col - X.shape[1]))))

    elif X.shape[1] > n_col:
        if sparse.issparse(X):
            X = X.tocsc()[:, :-(X.shape[1] - n_col)]
            X = X.tocsr()
        else:
            X = X[:, :-(X.shape[1] - n_col)]

    return X


def rank(x):
    """Rank a vector x.  Ties will be averaged."""

    unique, idx_inverse = np.unique(x, return_inverse=True)

    unique_rank_sum = np.zeros_like(unique)
    unique_rank_count = np.zeros_like(unique)

    np.add.at(unique_rank_sum, idx_inverse, x.argsort().argsort())
    np.add.at(unique_rank_count, idx_inverse, 1)

    unique_rank_mean = unique_rank_sum.astype(np.float) / unique_rank_count

    return unique_rank_mean[idx_inverse]


def set_min_max(x, lb, ub):
    x[x < lb] = lb
    x[x > ub] = ub

    return x


def point(rank, n_team, n_teammate=1, t=0):
    """Calculate Kaggle points to earn after a competition.

    Args:
        rank (int): final ranking in the private leaderboard.
        n_team (int): the number of teams participated in the competition.
        n_teammate (int): the number of team members in my team.
        t (int): the number of days since the competition ends.

    Returns:
        returns Kaggle points to earn after a compeittion.
    """
    return (1e5 / np.sqrt(n_teammate) * (rank ** -.75) *
            np.log10(1 + np.log10(n_team)) * np.exp(-t / 500))



from setuptools import setup, Extension
from Cython.Distutils import build_ext

import numpy as np

try:
    from pypandoc import convert
    read_md = lambda f: convert(f, 'rst')
except ImportError:
    print("warning: pypandoc module not found, could not convert Markdown to RST")
    read_md = lambda f: open(f, 'r').read()

setup(
    name='Kaggler',
    version='0.5.0',

    author='Jeong-Yoon Lee',
    author_email='jeongyoon.lee1@gmail.com',

    packages=['kaggler',
              'kaggler.feature_selection',
              'kaggler.model',
              'kaggler.metrics',
              'kaggler.online_model',
              'kaggler.preprocessing',
              'kaggler.test'],
    url='https://github.com/jeongyoonlee/Kaggler',
    license='LICENSE.txt',

    description='Code for Kaggle Data Science Competitions.',
    long_description=read_md('README.md'),

    install_requires=[
        'cython',
        'h5py',
        'ml_metrics',
        'numpy',
        'pandas',
        'scipy >= 0.14.0',
        'scikit-learn >= 0.15.0',
        'statsmodels >= 0.5.0',
    ],

    cmdclass={'build_ext': build_ext},
    ext_modules=[Extension('kaggler.online_model.ftrl',
                           ['kaggler/online_model/ftrl.pyx',
                            'kaggler/online_model/murmurhash/MurmurHash3.cpp'],
                           libraries=[],
                           include_dirs=[np.get_include(), '.'],
                           extra_compile_args=['-O3']),
                 Extension('kaggler.online_model.sgd',
                           ['kaggler/online_model/sgd.pyx'],
                           libraries=[],
                           include_dirs=[np.get_include(), '.'],
                           extra_compile_args=['-O3']),
                 Extension('kaggler.online_model.fm',
                           ['kaggler/online_model/fm.pyx'],
                           libraries=[],
                           include_dirs=[np.get_include(), '.'],
                           extra_compile_args=['-O3']),
                 Extension('kaggler.online_model.nn',
                           ['kaggler/online_model/nn.pyx'],
                           libraries=[],
                           include_dirs=[np.get_include(), '.'],
                           extra_compile_args=['-O3']),
                 Extension('kaggler.online_model.nn_h2',
                           ['kaggler/online_model/nn_h2.pyx'],
                           libraries=[],
                           include_dirs=[np.get_include(), '.'],
                           extra_compile_args=['-O3']),
                 Extension('kaggler.util',
                           ['kaggler/util.pyx', 'kaggler/util.pxd'],
                           libraries=[],
                           include_dirs=[np.get_include(), '.'],
                           extra_compile_args=['-O3'])],
)



import numpy as np
import os
import unittest

from kaggler.preprocessing import Normalizer, LabelEncoder, OneHotEncoder


DUMMY_X1 = np.array([[3, 'a'],
                     [3, 'b'],
                     [1, 'a'],
                     [np.nan, 'c'],
                     [np.nan, 'b'],
                     [0, 'c']])

class TestLabelEncoder(unittest.TestCase):

    def setUp(self):
        self.encoder = LabelEncoder(min_obs=2)

    def test_fit(self):
        self.encoder.fit(DUMMY_X1)
            
        # check if target values are correct
        self.assertEqual(self.encoder.label_maxes, [2, 3])


if __name__ == '__main__':
    unittest.main()




import os
import unittest

from kaggler.online_model import SGD


DUMMY_SPARSE_STR = """0 1:1 3:1 10:1
0 3:1 5:1
1 4:1 6:1 8:1 10:1"""

DUMMY_Y = [0, 0, 1]
DUMMY_LEN_X = [3, 2, 4]

class TestSGD(unittest.TestCase):

    def setUp(self):
        self.model = SGD(n=2**10, a=0.1, l1=1, l2=1, interaction=True)
        self.sparse_file = '/tmp/dummy.sps'

        """Create dummpy sparse files."""
        with open(self.sparse_file, 'w') as f:
            f.write(DUMMY_SPARSE_STR)

    def tearDown(self):
        # If a dummy file exists, remove it.
        if os.path.isfile(self.sparse_file):
            os.remove(self.sparse_file)

    def test_read_sparse(self):
        len_xs = []
        ys = []
        for x, y in self.model.read_sparse(self.sparse_file):
            # check hash collision for feature index
            self.assertEqual(len(set(x)), len(x))

            ys.append(y)
            len_xs.append(len(x))
            
        # check if target values are correct
        self.assertEqual(ys, DUMMY_Y)

        # check if the number of feature index are correct
        self.assertEqual(len_xs, DUMMY_LEN_X)


if __name__ == '__main__':
    unittest.main()

